function OutStruct = aivga(dataset, varargin)
% AIVGA Agglomerative Independent Variable Group Analysis
%
%  res = AIVGA(D, ['arg1', val1, 'arg2', val2, ...])
%    returns a hierarchical grouping of the variables in the dataset D, where
%   res.costs are the cost function values at each state
%   res.moves has the indices of groups joined at each state in its columns
%   res.groupings holds the groupings at each level of the hierarchy
%   res.models has compressed representations of the models from each step
% 
% The following optional keyword arguments are accepted:
%    'mode'       is the operation mode, possible values are:
%                 'symm' (symmetric algorithm, default) and
%                 'asymm' (asymmetric algorithm).
%                 With the asymmetric algorithm, the first variable in
%                 D is assumed to be the class or regressor variable
%                 and only moves involving it are allowed.
%    'impl_noise' is the variance of the implicit noise (default: 0)
%    'update_hyperparams'
%                 specifies if the hyperparameters are updated in
%                 DP mixture modeller                   (default: no = 0)
%    'mixtalg'    is the mixture algorithm, possible values are
%                 'dp' (Dirichlet process mixtures, default) and
%                 'classic' (finite mixture model, deprecated).
%    'maxiters'   is the maximum number of iterations in classic mixture model
%                 training                              (default: 1000)
%    'initcomps'  is the initial number of mixture components for
%                 new classic models                    (default: 2)
%    'maxcomps'   is the maximum number of mixture components for
%                 classic models                        (default: 100)
%
% The dataset D can be either a structure returned by AIVGA_IMPORT_DATA,
% or anything accepted by AIVGA_IMPORT_DATA as a single argument. In the
% latter case, AIVGA_IMPORT_DATA is applied to the argument.
%
% (c) 2006-2007 Esa Alhoniemi, Antti Honkela and Jeremias Seppa

% This file is a part of the AIVGA program
%
% Copyright (C) 2006-2007 Esa Alhoniemi, Antti Honkela and Jeremias Seppa
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 2, or (at your option)
% any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License (included in file License.txt in the
% program package) for more details.


% Read the arguments
if (length(varargin) == 1) & (isstruct(varargin{1})),
  args = varargin{1};
elseif (mod(length(varargin), 2) ~= 0),
  if ~(isstruct(varargin{1})),
    error('Keyword arguments should appear in pairs');
  else
    args = varargin{1};
    for k=2:2:length(varargin),
      if ~ischar(varargin{k})
        error('Keyword argument names must be strings.');
      end
      eval(sprintf('args.%s = varargin{k+1};', varargin{k}));
    end
  end
else
  for k=1:length(varargin),
    if iscell(varargin{k}) && numel(varargin{k}) > 1,
      varargin{k} = {varargin{k}};
    end
  end
  args = struct(varargin{:});
end

% Get the values
[params.mode, args] = getargdef(args, 'mode', 'symm');
[params.maxiter, args] = getargdef(args, 'maxiters', 1000);
[params.implicit_noisevar, args] = getargdef(args, 'impl_noise', 0);
[params.update_hyperparams, args] = getargdef(args, 'update_hyperparams', 0);
[params.initcomps, args] = getargdef(args, 'initcomps', 2);
[params.maxcomps, args] = getargdef(args, 'maxcomps', 100);
[params.mixtalg, args] = getargdef(args, 'mixtalg', 'dp');

% Report unused arguments
otherargs = fieldnames(args);
if length(otherargs) > 0,
  fprintf('Warning: aivga: unused arguments:\n');
  fprintf(' %s\n', otherargs{:});
end

if exist('vdp_mk_log_lambda', 'file') == 3 && ...
      exist('vdp_mk_hp_posterior', 'file') == 3 && ...
      exist('vdp_sumlogsumexp', 'file') == 3 && ...
      exist('vdp_softmax', 'file') == 3,
  params.usemex = 1;
else
  warning('MEX files not found, try using COMPILE to create them and speed up the algorithm a lot');
  params.usemex = 0;
end

if ~isstruct(dataset),
  dataset = aivga_import_data(dataset);
end

[N,dim] = size(dataset.data);
dim0    = dim;

%
% compute cost for each variable
%

C = 0;
H = zeros(1, dim);
for k = 1:dim
  [H(k) m]     = cost_a(dataset, k, params);
  C            = C + H(k);
  modelhist{k} = m;
  fprintf('\rComputing model for variable %4d/%4d...', k, dim);
end
fprintf('done.\n');

switch params.mode
 case 'symm'
  delta   = zeros(dim);
  models  = cell(dim);

  %
  % compute costs for the combined groups
  %

  for a = 1:dim
    fprintf('\rComputing delta values for variable %4d/%4d...', a, dim);
    for b = a+1:dim
      [cost m]     = cost_a(dataset, [a b], params);
      delta(a, b)  = cost - H(a) - H(b);     % change (increase) of the total cost
      delta(b, a)  = delta(a, b);
      models{a, b} = m;
      models{b, a} = m;
    end
    G{a}        = a;
    delta(a, a) = Inf;
  end
  fprintf('done.\n');

  move_cost_hist  = [0; 0; C];
  groupinghist    = {};   
  groupinghist{1} = {G};   

  for j = 2:dim0
    fprintf('\rCombining groups, %4d group(s) left...', dim - 2);
  
    m = inf;
    for i = 1:dim
      [x z] = min(delta(i, :));
      if (x < m)
	m = x;
	b = z;            % we search for the indices (a,b) of the groups
	a = i;            % that are joined next ..
      end
    end
    C = C + m; % oli: x
    move_cost_hist      = [move_cost_hist [a; b; C]];
    modelhist{dim0 +j-1} = models{a,b};
    temp                = sort([a b]);
    a                   = temp(1);  % this sorting maybe not needed? or?
    b                   = temp(2);
    new_group_cost      = m + H(a) + H(b); % oli: x
    removed_group_vars  = G{b};

    %
    % remove group b
    %

    G           = [G(1:b-1) G(b+1:dim)];
    H           = [H(1:b-1) H(b+1:dim)];
    delta(:, b) = [];
    delta(b, :) = [];
    dim         = dim - 1;

    %
    % put the new group to a's place
    %
    
    G{a} = sort([G{a} removed_group_vars]);
    H(a) = new_group_cost;

    groupinghist{j} = G;
    for i = [1:a-1 a+1:dim]  % not run at last iteration, update "distances"
      [cost, model] = cost_a(dataset, sort([G{a} G{i}]), params);  
      cdiff         = cost - H(a) - H(i);
      delta(a,i)    = cdiff;
      delta(i,a)    = cdiff;
      models{a,i}   = model;
      models{i,a}   = model;
    end
  end
 case 'asymm'
  delta   = zeros(dim, 1);
  models  = cell(dim, 1);

  %
  % compute costs for the combined groups
  %

  a = 1;
  G{a} = a;
  for b = a+1:dim
    fprintf('\rComputing delta values for variable %4d/%4d...', b, dim);
    [cost m]  = cost_a(dataset, [a b], params);
    delta(b)  = cost - H(a) - H(b);     % change (increase) of the total cost
    models{b} = m;
    G{b}      = b;
  end
  delta(a) = Inf;

  fprintf('done.\n');

  move_cost_hist  = [0; 0; C];
  groupinghist    = {};   
  groupinghist{1} = {G};   

  for l = 2:dim0
    fprintf('\rCombining groups, %4d group(s) left...', dim - 2);

    a = 1;
    [m b] = min(delta);

    C = C + m; % oli: x
    move_cost_hist      = [move_cost_hist [a; b; C]];
    modelhist{dim0+l-1} = models{b};
    temp                = sort([a b]);
    a                   = temp(1);  % this sorting maybe not needed? or?
    b                   = temp(2);
    new_group_cost      = m + H(a) + H(b); % oli: x
    removed_group_vars  = G{b};

    %
    % remove group b
    %

    G           = [G(1:b-1) G(b+1:dim)];
    H           = [H(1:b-1) H(b+1:dim)];
    delta(b)    = [];
    dim         = dim - 1;

    %
    % put the new group to a's place
    %
  
    G{a} = sort([G{a} removed_group_vars]);
    H(a) = new_group_cost;

    groupinghist{l} = G;
    for k = a+1:dim  % not run at last iteration, update "distances"
      [cost, model] = cost_a(dataset, sort([G{a} G{k}]), params);  
      cdiff         = cost - H(a) - H(k);
      delta(k)      = cdiff;
      models{k}     = model;
    end
  end
 otherwise
  error('Unknown mode')
end

fprintf('done.\n');

OutStruct.cost=move_cost_hist(3,:);
OutStruct.moves=move_cost_hist(1:2,:);
OutStruct.groupings=groupinghist;
OutStruct.models=modelhist;



function [val, args] = getargdef(args, name, default),

if isfield(args, name),
  % val = args.(name);
  eval(sprintf('val = args.%s;', name));
  args = rmfield(args, name);
else
  val = default;
end



function [cost, modelpack] = cost_a(D, vars, params)

switch params.mixtalg,
 case 'dp'
  newd = D;
  newd.data = newd.data(:, vars);
  newd.types = newd.types(vars);
  newd.S = newd.S(vars);
  newd.X1 = newd.data(:, (newd.types == 1));
  newd.X2 = newd.data(:, (newd.types == 2));
  newd.realS = newd.S(newd.types == 2);
  res = vdp_mixt(newd, struct('update_hyperparams', params.update_hyperparams, 'implicit_noisevar', params.implicit_noisevar, 'usemex', params.usemex));
 
  cost = res.free_energy;
  modelpack = res;
 
 case 'classic'
  [cost, modelpack] = finite_mixt(D, vars, params);

 otherwise
  error('Unknown mixture algoritm');
end
