function [cost, modelpack] = finite_mixt(D, vars, params)
% [cost, model] = FINITE_MIXT(D, VARS, PARAMS)

% This file is a part of the AIVGA program
%
% Copyright (C) 2006-2007 Esa Alhoniemi, Antti Honkela and Jeremias Sepp
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 2, or (at your option)
% any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License (included in file License.txt in the
% program package) for more details.

% randomly select new number of mixture components (=="centroids") for
% the group and compute a new random seed; try this reinitialization a
% few times

ncentroids   = params.initcomps;   % initial number of centroids
maxcentroids = params.maxcomps;
l            = length(vars);
max_reinits  = 12 * floor(sqrt(l) * 1.3 +.001); 
% was earlier   6                   1.2

if (l == 1 & D.types(vars) == 2) % minor optimization
  ncentroids  = 1;
  max_reinits = 2;
end

cost = Inf;

for n = 1:max_reinits
  ncentroids_ = min(max(round(ncentroids*exp(2*(rand(1)-0.4))),1),maxcentroids);
  cost_       = NaN;
  
  % run a new vq

  while (isnan(cost_))
    seed_                  = sum(100*clock);
    [cost_ iters_] = avq( D, vars, ncentroids_, seed_, cost, params);
  end

  % if cost improved, store it

  if cost_ < cost
    cost       = cost_;    
    ncentroids = ncentroids_;
    seed       = seed_;
    iters      = iters_;   
  end
end

modelpack = [ncentroids seed iters vars];


function [cost iters] = avq(D, Vars, ncentroids, seed, record, params)

% record = the best cost so far for this variable group

%abs_eps_a=3e-2;   % old values
%rel_eps_a=5e-4;
%abs_eps_b=2e-3;   
%rel_eps_b=2e-6;

maxiter   = params.maxiter;
step      = 12;

implicit_noisevar = params.implicit_noisevar;

abs_eps_a = 2e-2;   
rel_eps_a = 2e-4;
abs_eps_b = 1e-3;   % used when "making a new record cost"
rel_eps_b = 1e-6;

iters     = 0;
verbose   = 0;
params    = [ncentroids seed];

X         = D.data(:, Vars);
types     = D.types(Vars);
S         = D.S(Vars);
nnom      = sum(types == 2);
last_cost = Inf;
v         = length(Vars);
abs_eps   = abs_eps_a;
rel_eps   = rel_eps_a;

while(1)
  params = vectquantM(X, types, S, params, step, verbose, implicit_noisevar);
  iters  = iters + step;
  
  params = remove_empty(params, nnom);

  params = vectquantM(X, types, S, params, 3, verbose, implicit_noisevar);
  iters  = iters + 3;

  cost   = vq_cost(D, params, Vars);

  if (cost < record)
    abs_eps = abs_eps_b;
    rel_eps = rel_eps_b;
  end

  % end condition
  if (((last_cost/cost) < (1+rel_eps)) || ...
      ((last_cost - cost) < abs_eps) || (iters>maxiter))
    break; 
  end

  last_cost = cost;
end

return;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [Params, ncentroids] = remove_empty(Params, nnom)

Inds = find(Params.F < 0.5);

Params.Mu_tilde(Inds,:)  = []; 
Params.Mu_bar(Inds,:)    = [];
Params.Ksi_alpha(Inds,:) = [];
Params.Ksi_beta(Inds,:)  = [];
Params.F(Inds)           = [];
Params.W(Inds,:)         = [];
Params.logC(Inds)        = [];

if nnom
 for ind2 = 1:nnom
   Params.U_hat{ind2}(Inds,:) = [];
 end
end

ncentroids = length(Params.F);


function Params = vectquantM(X, types, S, Init, max_iters, verbose, implicit_noisevar)

% Params = vectquantM(X, types, S,[ncentroids seed], max_iters, verbose)
% Params = vectquantM(X, types, S,Init, max_iters, verbose)
%
% INPUT ARGUMENTS
%
%
%   ncentroids   number of codebook vectors
%   seed         seed number for random number generator
%   Init         struct consisting of model parameters, for example 
%                output argument Params from some previous run
%   [max_iters]  maximum number of iterations, default = 100
%                if max_iters == 1, it will be silently raised to 2.   (*
%   verbose      currently has no effect.
%
% OUTPUT ARGUMENTS
%  
%   Params       struct consisting of model parameters
%


%  This file is a part of the Independent Variable Group Analysis package
% 
%  Copyright (C) 2001-2006 Esa Alhoniemi, Antti Honkela, Krista Lagus,
%  Jeremias Seppa, Harri Valpola, and Paul Wagner
% 
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
% 
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License (included in file License.txt in the
%  program package) for more details.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

X1       = X(:, (types == 1)); % real-valued
X2       = X(:, (types == 2)); % nominal
S        = S(types == 2);
[n dim1] = size(X1);
dim2     = size(X2, 2);

old_rand_state = [];
if max_iters == 1, max_iters = 2; end

%
% initialize
%

if isstruct(Init)
     
  % continue earlier run

else
  % make a new vq
  
  ncentroids    = Init(1);
  seed          = Init(2);
  
  Mean          = sum(X1, 1)/n;
  Var           = sum((X1 - repmat(Mean,n,1)).^2, 1)/n; % variance
  
  % priors for distribution of codebook vectors Mu ~ N(Mu_Mu, S2_Mu)..

  Init = struct('Mu_mu', Mean, 'S2_mu', Var);
  if dim2
    for j = 1:dim2
      Init.U_p(j) = 0.5;
    end
  else
    Init.U_p = Inf;
  end

  % posteriors for distribution of codebook vectors Mu ~ N(Mu_bar, Mu_tilde)..

  old_rand_state = rand('state');
  rand('state', seed);
  Inds          = randperm(n);
  Init.Mu_bar   = X1(Inds(1:ncentroids), :);
  Init.Mu_tilde = repmat(1e-100, ncentroids, dim1);
  Uhat = cell(1, dim2);
  for j = 1:dim2
    Uhat{j} = 0.7+0.5*rand(ncentroids, S(j));
    for i = 1:ncentroids
      k            = X2(Inds(i), j);
      Uhat{j}(i,k) = Uhat{j}(i,k) + 3;
    end
  end  
  Init.U_hat = Uhat;
  
  % priors for data variance Ksi ~ Gamma(Alpha_ksi, Beta_ksi)
  
  Init.Alpha_ksi = ones(1, dim1);
  Init.Beta_ksi  = Var/ncentroids;

  % posteriors for data variance Ksi ~ Gamma(Ksi_alpha, Ksi_beta)

  Init.Ksi_alpha = ones(ncentroids, dim1);
  Init.Ksi_beta  = repmat((Var/ncentroids), ncentroids, 1);
  
  % winners for data points C ~ Dirichlet(F) (F contains winner frequencies)
  
  Init.C_prior  = 0.5;
  Init.logC     = log(ones(ncentroids, 1)/ncentroids);
end

Init.Mu_bar    = Init.Mu_bar';
Init.Mu_tilde  = Init.Mu_tilde';
Init.Ksi_alpha = Init.Ksi_alpha';
Init.Ksi_beta  = Init.Ksi_beta';

Params = mix(X1', Init, max_iters, verbose, X2, S, implicit_noisevar);

Params.Mu_bar    = Params.Mu_bar';
Params.Mu_tilde  = Params.Mu_tilde';
Params.Ksi_alpha = Params.Ksi_alpha';
Params.Ksi_beta  = Params.Ksi_beta';

if ~isempty(old_rand_state), rand( 'state', old_rand_state ); end

return;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



function cost = vq_cost(D, Params, Vars)

% Compute the cost using given model parameters and data
% Author: Jeremias Seppa

%if isstruct(D.data)  % 'auxiliary invocation' for 
%  X1=D.data.data1;   % e.g. debugging purposes. activate
%  X2=D.data.data2;   % by removing comment marks
%  S=D.S;
%else
 types=D.types(Vars);
 X=D.data(:,Vars);
 S=D.S(Vars);
 X1=X(:,(types==1));
 X2=X(:,(types==2));
 S=S(types==2);
%end

Uhat = Params.U_hat;
U_p = Params.U_p;
Mu_mu = Params.Mu_mu;
S2_mu = Params.S2_mu;
Mu_bar = Params.Mu_bar;
Mu_tilde = Params.Mu_tilde;
Alpha_ksi = Params.Alpha_ksi;
Beta_ksi = Params.Beta_ksi;
Ksi_alpha = Params.Ksi_alpha;
Ksi_beta = Params.Ksi_beta;
c_p = Params.C_prior;
logc = Params.logC;
F = Params.F;
W = Params.W;
Ksi_log   = (psi(Ksi_alpha)-log(Ksi_beta));
S2_x      = Ksi_beta./Ksi_alpha;

[N,dim1]=size(X1);
dim2=size(X2,2);
C=size(W,1);

Mu_mu     = repmat(Mu_mu,C,1);
S2_mu     = repmat(S2_mu,C,1);
Alpha_ksi = repmat(Alpha_ksi,C,1);
Beta_ksi  = repmat(Beta_ksi,C,1);

l_codebook = C*(gammaln(c_p)) - dim1*C/2;

for j=1:dim2,
  l_codebook=l_codebook - (C*gammaln(U_p(j)*S(j))) + ...
      C*S(j)*gammaln(U_p(j)) + ...
      sum(gammaln(sum(Uhat{j},2))) - sum(sum(gammaln(Uhat{j})));
  l_codebook=l_codebook + sum(sum((Uhat{j}-U_p(j)) .* ...
				  (psi(Uhat{j}) - ...
				   repmat(psi(sum(Uhat{j},2)),1,S(j)))));
end

if dim1
 l_codebook=l_codebook + ...
   sum(sum(((log(S2_mu./Mu_tilde)+ ...
	     (Mu_tilde+(Mu_bar-Mu_mu).^2)./S2_mu)/2)));

 l_codebook=l_codebook + ...
   sum(sum(gammaln(Alpha_ksi)-gammaln(Ksi_alpha)+Ksi_alpha.*log(Ksi_beta) ...
	   - Alpha_ksi .* log(Beta_ksi) + (Ksi_alpha - Alpha_ksi) .* ...
	   Ksi_log + (Beta_ksi - Ksi_beta) .* Ksi_alpha./Ksi_beta));
end
l_codebook=l_codebook - gammaln(C*c_p) + sum(F.*logc - gammaln(F+c_p)) +...
    gammaln((C*c_p)+sum(F));
  
l_winners=sum(sum(W.*log(W+(1e-100)))) + sum(-F .* logc);

temp=zeros(C,N);
for j=1:dim2,
  UhatPsi=psi(Uhat{j});
  temp=temp - UhatPsi(:,X2(:,j))+repmat(psi(sum(Uhat{j},2)),1,N);
end
for j=1:dim1
  for k=1:C
    temp(k, :) = temp(k, :) + 0.5*((Mu_tilde(k,j)+ ...
				    (X1(:,j)'- Mu_bar(k,j)).^2) ...
				   ./ S2_x(k, j) ...
				   - Ksi_log(k, j));
  end
end
% (or forloop over t instead of j if for is fast enough - less repmats)

l_errors = dim1*N*log(2*pi)/2 + sum(sum(W .* temp));

cost = l_codebook + l_winners + l_errors;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

